/*
 *  filefuncs.cpp - helper functions to deal with files platform independently
 *
 *  Copyright (C) 2002-2003 Glen Sawyer & Marcel Mueller
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Ported to ANSI C (or at least somethingvery similar) by Marcel Mller.
 */


#include "filefuncs.h"

#ifdef WIN32
   #include <windows.h>
#endif

#include <stdio.h>
#include <assert.h>
#include <utime.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(__BEOS__)
   #include <fs_attr.h>
   #include <bsd_mem.h>
#endif
#include <fcntl.h>


long getSizeOfFile(const char *filename)
{  long size = 0;
   FILE *file;

   file = fopen(filename, "rb");
   if (file)
   {  fseek(file, 0, SEEK_END);
      size = ftell(file);
      fclose(file);
   }

   return size;
}

void fileTime(const char *filename, timeAction action)
{
   static bool timeSaved = false;
#ifdef WIN32
   HANDLE outfh;
   static FILETIME create_time, access_time, write_time;
#else
   static struct stat savedAttributes;
#endif

   if (action == storeTime) {
#ifdef WIN32
      outfh = CreateFile((LPCTSTR)filename,
                     GENERIC_READ,
                     FILE_SHARE_READ,
                     NULL,
                     OPEN_EXISTING,
                     FILE_ATTRIBUTE_NORMAL,
                     NULL);
      if (outfh != INVALID_HANDLE_VALUE)
      {  if (GetFileTime(outfh,&create_time,&access_time,&write_time))
            timeSaved = true;
         CloseHandle(outfh);
      }
#else
      timeSaved = stat(filename, &savedAttributes) == 0;
#endif
   } else if (timeSaved) {
#ifdef WIN32
      outfh = CreateFile((LPCTSTR)filename,
               GENERIC_WRITE,
               0,
               NULL,
               OPEN_EXISTING,
               FILE_ATTRIBUTE_NORMAL,
               NULL);
      if (outfh != INVALID_HANDLE_VALUE)
      {  SetFileTime(outfh,&create_time,&access_time,&write_time);
         CloseHandle(outfh);
      }
#else
      struct utimbuf setTime;

      setTime.actime = savedAttributes.st_atime;
      setTime.modtime = savedAttributes.st_mtime;
      timeSaved = false;

      utime(filename, &setTime);
#endif
   }
}

void fileCopyAttributes(FILE* inf, FILE* out)
{
#ifdef __BEOS__
   /* some stuff to preserve attributes */
   do {
       DIR *attrs = NULL;
       struct dirent *de;
       struct attr_info ai;
       int infd, outfd;
       void *attrdata;

       infd = fileno(inf);
       if (infd < 0)
           goto attrerror;
       outfd = fileno(outf);
       if (outfd < 0)
           goto attrerror;
       attrs = fs_fopen_attr_dir(infd);
       while ((de = fs_read_attr_dir(attrs)) != NULL) {
           if (fs_stat_attr(infd, de->d_name, &ai) < B_OK)
               goto attrerror;
           if ((attrdata = malloc(ai.size)) == NULL)
               goto attrerror;
           fs_read_attr(infd, de->d_name, ai.type, 0, attrdata, ai.size);
           fs_write_attr(outfd, de->d_name, ai.type, 0, attrdata, ai.size);
           free(attrdata);
       }
       fs_close_attr_dir(attrs);
       break;
   attrerror:
       if (attrdata)
           free(attrdata);
       if (attrs)
           fs_close_attr_dir(attrs);
       fprintf(stderr, "can't preserve attributes for '%s': %s\n", filename, strerror(errno));
   } while (0);
#endif
// We should do some other platform specific attribute stuff here.
// e.g. extended attributes on OS/2, Win32 and I guess AIX, too.
}

