/*
 *  ReplayGainAnalysis - analyzes input samples and give the recommended dB change
 *  Copyright (C) 2001-2002 David Robinson and Glen Sawyer & Marcel Mueller
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  concept and filter values by David Robinson (David@Robinson.org)
 *    -- blame him if you think the idea is flawed
 *  original coding by Glen Sawyer (glensawyer@hotmail.com)
 *    -- blame him if you think this runs too slowly, or the coding is otherwise flawed
 *
 *  lots of code improvements by Frank Klemm ( http://www.uni-jena.de/~pfk/mpp/ )
 *    -- credit him for all the _good_ programming ;)
 *
 *
 *  For an explanation of the concepts and the basic algorithms involved, go to:
 *    http://www.replaygain.org/
 */

/*
 *  So here's the main source of potential code confusion:
 *
 *  The filters applied to the incoming samples are IIR filters,
 *  meaning they rely on up to <filter order> number of previous samples
 *  AND up to <filter order> number of previous filtered samples.
 *
 *  I set up the AnalyzeSamples routine to minimize memory usage and interface
 *  complexity. The speed isn't compromised too much (I don't think), but the
 *  internal complexity is higher than it should be for such a relatively
 *  simple routine.
 *
 *  Optimization/clarity suggestions are welcome.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <time.h>

#include "gain_analysis.h"

typedef unsigned short  Uint16_t;
typedef signed short    Int16_t;
typedef unsigned int    Uint32_t;
typedef signed int      Int32_t;

// Systematic constants
#define YULE_ORDER         10
#define BUTTER_ORDER        2
#define YULE_FILTER     filter10
#define BUTTER_FILTER   filter2
// constant parameters
#define MAX_SAMP_FREQ   48000.          // maximum allowed sample frequency [Hz]
#define MAX_RMS_WINDOW_TIME 0.5         // maximum time slice size [s]
#define MAX_SAMPLES_PER_WINDOW  (size_t) (MAX_SAMP_FREQ * MAX_RMS_WINDOW_TIME)  // max. Samples per Time slice
#define MAX_ORDER               (BUTTER_ORDER > YULE_ORDER ? BUTTER_ORDER : YULE_ORDER)
#define STEPS_per_dB       20.          // Table entries per dB
#define MAX_dB            120.          // Table entries for 0...MAX_dB (normal max. values are 70...80 dB)
// default values
#define DEF_PINK_REF       64.82//298640883795 // calibration value
#define DEF_RMS_PERCENTILE  0.95        // percentile which is louder than the proposed level
#define DEF_RMS_WINDOW_TIME 0.1         // default time slice size [s]
// dynamic parameters
float                pink_ref;
float                rms_percentile;
float                rms_window_time;

// internal buffers
static Float_t       linprebuf [MAX_ORDER * 2];
static Float_t       rinprebuf [MAX_ORDER * 2];
static Float_t*      linpre;                                          // left input samples, with pre-buffer
static Float_t*      rinpre;                                          // right input samples ...
static double        lstepbuf  [MAX_SAMPLES_PER_WINDOW + MAX_ORDER];
static double        rstepbuf  [MAX_SAMPLES_PER_WINDOW + MAX_ORDER];
static double*       lstep;                                           // left "first step" (i.e. post first filter) samples
static double*       rstep;
static double        loutbuf   [MAX_SAMPLES_PER_WINDOW + MAX_ORDER];
static double        routbuf   [MAX_SAMPLES_PER_WINDOW + MAX_ORDER];
static double*       lout;                                            // left "out" (i.e. post second filter) samples
static double*       rout;
static unsigned int  sampleWindow;                                    // _half_ number of samples required to reach number of milliseconds required for RMS window
static unsigned long totsamp;
static double        sum1, sum2;                                      // sum of squares of the left and right window half
                                                                      // This is _not_ the left and right channel. There is no need to seperate them here.
static int           freqindex;
static int           first;                                           // first window half
static Uint32_t  A [(size_t)(STEPS_per_dB * MAX_dB)];
static Uint32_t  B [(size_t)(STEPS_per_dB * MAX_dB)];
static Float_t   window    [MAX_SAMPLES_PER_WINDOW];                  // buffer for squared window function

#ifdef PROFILE
static clock_t t0;
#endif

// for each filter:
// [0] 48 kHz, [1] 44.1 kHz, [2] 32 kHz, [3] 24 kHz, [4] 22050 Hz, [5] 16 kHz, [6] 12 kHz, [7] is 11025 Hz, [8] 8 kHz

#pragma warning ( disable : 4305 )

static const double ABYule[9][2*YULE_ORDER + 1] = {
    {0.03857599435200, -3.84664617118067, -0.02160367184185,  7.81501653005538, -0.00123395316851,-11.34170355132042, -0.00009291677959, 13.05504219327545, -0.01655260341619,-12.28759895145294,  0.02161526843274,  9.48293806319790, -0.02074045215285, -5.87257861775999,  0.00594298065125,  2.75465861874613,  0.00306428023191, -0.86984376593551,  0.00012025322027,  0.13919314567432,  0.00288463683916 },
    {0.05418656406430, -3.47845948550071, -0.02911007808948,  6.36317777566148, -0.00848709379851, -8.54751527471874, -0.00851165645469,  9.47693607801280, -0.00834990904936, -8.81498681370155,  0.02245293253339,  6.85401540936998, -0.02596338512915, -4.39470996079559,  0.01624864962975,  2.19611684890774, -0.00240879051584, -0.75104302451432,  0.00674613682247,  0.13149317958808, -0.00187763777362 },
    {0.15457299681924, -2.37898834973084, -0.09331049056315,  2.84868151156327, -0.06247880153653, -2.64577170229825,  0.02163541888798,  2.23697657451713, -0.05588393329856, -1.67148153367602,  0.04781476674921,  1.00595954808547,  0.00222312597743, -0.45953458054983,  0.03174092540049,  0.16378164858596, -0.01390589421898, -0.05032077717131,  0.00651420667831,  0.02347897407020, -0.00881362733839 },
    {0.30296907319327, -1.61273165137247, -0.22613988682123,  1.07977492259970, -0.08587323730772, -0.25656257754070,  0.03282930172664, -0.16276719120440, -0.00915702933434, -0.22638893773906, -0.02364141202522,  0.39120800788284, -0.00584456039913, -0.22138138954925,  0.06276101321749,  0.04500235387352, -0.00000828086748,  0.02005851806501,  0.00205861885564,  0.00302439095741, -0.02950134983287 },
    {0.33642304856132, -1.49858979367799, -0.25572241425570,  0.87350271418188, -0.11828570177555,  0.12205022308084,  0.11921148675203, -0.80774944671438, -0.07834489609479,  0.47854794562326, -0.00469977914380, -0.12453458140019, -0.00589500224440, -0.04067510197014,  0.05724228140351,  0.08333755284107,  0.00832043980773, -0.04237348025746, -0.01635381384540,  0.02977207319925, -0.01760176568150 },
    {0.44915256608450, -0.62820619233671, -0.14351757464547,  0.29661783706366, -0.22784394429749, -0.37256372942400, -0.01419140100551,  0.00213767857124,  0.04078262797139, -0.42029820170918, -0.12398163381748,  0.22199650564824,  0.04097565135648,  0.00613424350682,  0.10478503600251,  0.06747620744683, -0.01863887810927,  0.05784820375801, -0.03193428438915,  0.03222754072173,  0.00541907748707 },
    {0.56619470757641, -1.04800335126349, -0.75464456939302,  0.29156311971249,  0.16242137742230, -0.26806001042947,  0.16744243493672,  0.00819999645858, -0.18901604199609,  0.45054734505008,  0.30931782841830, -0.33032403314006, -0.27562961986224,  0.06739368333110,  0.00647310677246, -0.04784254229033,  0.08647503780351,  0.01639907836189, -0.03788984554840,  0.01807364323573, -0.00588215443421 },
    {0.58100494960553, -0.51035327095184, -0.53174909058578, -0.31863563325245, -0.14289799034253, -0.20256413484477,  0.17520704835522,  0.14728154134330,  0.02377945217615,  0.38952639978999,  0.15558449135573, -0.23313271880868, -0.25344790059353, -0.05246019024463,  0.01628462406333, -0.02505961724053,  0.06920467763959,  0.02442357316099, -0.03721611395801,  0.01818801111503, -0.00749618797172 },
    {0.53648789255105, -0.25049871956020, -0.42163034350696, -0.43193942311114, -0.00275953611929, -0.03424681017675,  0.04267842219415, -0.04678328784242, -0.10214864179676,  0.26408300200955,  0.14590772289388,  0.15113130533216, -0.02459864859345, -0.17556493366449, -0.11202315195388, -0.18823009262115, -0.04060034127000,  0.05477720428674,  0.04788665548180,  0.04704409688120, -0.02217936801134 }
};

static const double ABButter[9][2*BUTTER_ORDER + 1] = {
    {0.98621192462708, -1.97223372919527, -1.97242384925416,  0.97261396931306,  0.98621192462708 },
    {0.98500175787242, -1.96977855582618, -1.97000351574484,  0.97022847566350,  0.98500175787242 },
    {0.97938932735214, -1.95835380975398, -1.95877865470428,  0.95920349965459,  0.97938932735214 },
    {0.97531843204928, -1.95002759149878, -1.95063686409857,  0.95124613669835,  0.97531843204928 },
    {0.97316523498161, -1.94561023566527, -1.94633046996323,  0.94705070426118,  0.97316523498161 },
    {0.96454515552826, -1.92783286977036, -1.92909031105652,  0.93034775234268,  0.96454515552826 },
    {0.96009142950541, -1.91858953033784, -1.92018285901082,  0.92177618768381,  0.96009142950541 },
    {0.95856916599601, -1.91542108074780, -1.91713833199203,  0.91885558323625,  0.95856916599601 },
    {0.94597685600279, -1.88903307939452, -1.89195371200558,  0.89487434461664,  0.94597685600279 }
};


#pragma warning ( default : 4305 )

// When calling this procedure, make sure that ip[-order] and op[-order] point to real data!

static void
filter10 (const Float_t* input, double* output, size_t nSamples, const double* kernel)
{
    while (nSamples--) {
        *output = input [0] * kernel[0]
         - output[-1] * kernel[1]
         + input [-1] * kernel[2]
         - output[-2] * kernel[3]
         + input [-2] * kernel[4]
         - output[-3] * kernel[5]
         + input [-3] * kernel[6]
         - output[-4] * kernel[7]
         + input [-4] * kernel[8]
         - output[-5] * kernel[9]
         + input [-5] * kernel[10]
         - output[-6] * kernel[11]
         + input [-6] * kernel[12]
         - output[-7] * kernel[13]
         + input [-7] * kernel[14]
         - output[-8] * kernel[15]
         + input [-8] * kernel[16]
         - output[-9] * kernel[17]
         + input [-9] * kernel[18]
         - output[-10]* kernel[19]
         + input [-10]* kernel[20];
        ++output;
        ++input;
    }
}

static void
filter2 (const double* input, double* output, size_t nSamples, const double* kernel)
{   while (nSamples--) {
        *output = input [0]  * kernel[0]
         - output[-1] * kernel[1]
         + input [-1] * kernel[2]
         - output[-2] * kernel[3]
         + input [-2] * kernel[4];
        ++output;
        ++input;
    }
}


// zero out the first num elements of buf
#define zerobuf(buf, num) memset((buf), 0, (num) * sizeof *(buf));

// reset gain sums for each title
static void resetGain ()
{
    #ifdef PROFILE
    t0 = clock();
    #endif

    // zero out initial values
    zerobuf(linprebuf, MAX_ORDER);
    zerobuf(rinprebuf, MAX_ORDER);
    zerobuf(lstepbuf, MAX_ORDER);
    zerobuf(rstepbuf, MAX_ORDER);
    zerobuf(loutbuf, MAX_ORDER);
    zerobuf(routbuf, MAX_ORDER);

    sum1 = sum2 = 0.;
    totsamp     = 0;
    first       = 1;

    memset ( A, 0, sizeof(A) );
}

// square
static inline double fsqr(const double d)
{  return d*d;
}

// add squared value to both window halfs (overlap!) and scale with the corresponding window function
static inline void windowadd(double sqr, Float_t win1, Float_t win2)
{   sum1 += sqr * win1;
    sum2 += sqr * win2;
}

// returns a INIT_GAIN_ANALYSIS_OK if successful, INIT_GAIN_ANALYSIS_ERROR if not

int ResetSampleFrequency ( long samplefreq ) {
    unsigned int i;

    switch ( (int)(samplefreq) ) {
        case 48000: freqindex = 0; break;
        case 44100: freqindex = 1; break;
        case 32000: freqindex = 2; break;
        case 24000: freqindex = 3; break;
        case 22050: freqindex = 4; break;
        case 16000: freqindex = 5; break;
        case 12000: freqindex = 6; break;
        case 11025: freqindex = 7; break;
        case  8000: freqindex = 8; break;
        default:    return INIT_GAIN_ANALYSIS_ERROR;
    }

    sampleWindow = (int)ceil(samplefreq * rms_window_time/2); // half window size for overlapping windows

    // Calculate _squared_ Hamming window function and integrate it numerically.
    Float_t* wp = window;
    double stepangle = 2 * M_PI / (2*sampleWindow - 1);
    double integral = 0.;
    for ( i = 0; i < 2*sampleWindow; ++i)
        integral += *wp++ = fsqr(.54 - .46 * cos(i * stepangle));
    // Normalize window funtion to meet sum(window) == 1.
    wp = window;
    integral = .5 / integral; // scale by 0.5 / sum (2 channels, also in mono mode!)
    for ( i = 0; i < 2*sampleWindow; ++i)
        *wp++ *= integral;

    #ifdef DEBUG // write window function
    wp = window;
    FILE* f = fopen("window.out", "w");
    for ( i = 0; i < 2*sampleWindow; ++i)
        fprintf(f, "%g\n", *wp++);
    fclose(f);
    #endif

    resetGain();

    return INIT_GAIN_ANALYSIS_OK;
}

int
InitGainAnalysis ( long samplefreq )
{   // init parameters
    pink_ref = DEF_PINK_REF;
    rms_percentile = DEF_RMS_PERCENTILE;
    rms_window_time = DEF_RMS_WINDOW_TIME;

    if (samplefreq && ResetSampleFrequency(samplefreq) != INIT_GAIN_ANALYSIS_OK)
       return INIT_GAIN_ANALYSIS_ERROR;

    linpre       = linprebuf + MAX_ORDER;
    rinpre       = rinprebuf + MAX_ORDER;
    lstep        = lstepbuf  + MAX_ORDER;
    rstep        = rstepbuf  + MAX_ORDER;
    lout         = loutbuf   + MAX_ORDER;
    rout         = routbuf   + MAX_ORDER;

    memset( B, 0, sizeof B );

    return INIT_GAIN_ANALYSIS_OK;
}

// returns GAIN_ANALYSIS_OK if successful, GAIN_ANALYSIS_ERROR if not

int
AnalyzeSamples ( const Float_t* left_samples, const Float_t* right_samples, size_t num_samples, int num_channels )
{
    const Float_t*  curleft;
    const Float_t*  curright;
    const Float_t*  curwindow1;
    const Float_t*  curwindow2;
    const double*   curleftB;
    const double*   currightB;
    unsigned long   batchsamples;
    unsigned long   cursamples;
    unsigned long   cursamplepos;
    int             i;

    if ( num_samples == 0 )
        return GAIN_ANALYSIS_OK;

    cursamplepos = 0;
    batchsamples = num_samples;

    switch ( num_channels) {
    case  1: right_samples = left_samples;
    case  2: break;
    default: return GAIN_ANALYSIS_ERROR;
    }

    if ( num_samples < MAX_ORDER ) {
        memcpy ( linprebuf + MAX_ORDER, left_samples , num_samples * sizeof *linprebuf );
        memcpy ( rinprebuf + MAX_ORDER, right_samples, num_samples * sizeof *rinprebuf );
    }
    else {
        memcpy ( linprebuf + MAX_ORDER, left_samples,  MAX_ORDER   * sizeof *linprebuf );
        memcpy ( rinprebuf + MAX_ORDER, right_samples, MAX_ORDER   * sizeof *rinprebuf );
    }

    while ( batchsamples > 0 ) {
        cursamples = batchsamples > sampleWindow-totsamp  ?  sampleWindow - totsamp  :  batchsamples;
        if ( cursamplepos < MAX_ORDER ) {
            curleft  = linpre+cursamplepos;
            curright = rinpre+cursamplepos;
            if (cursamples > MAX_ORDER - cursamplepos )
                cursamples = MAX_ORDER - cursamplepos;
        }
        else {
            curleft  = left_samples  + cursamplepos;
            curright = right_samples + cursamplepos;
        }

        YULE_FILTER ( curleft , lstep + totsamp, cursamples, ABYule[freqindex]);
        YULE_FILTER ( curright, rstep + totsamp, cursamples, ABYule[freqindex]);

        BUTTER_FILTER ( lstep + totsamp, lout + totsamp, cursamples, ABButter[freqindex]);
        BUTTER_FILTER ( rstep + totsamp, rout + totsamp, cursamples, ABButter[freqindex]);

        // Get the squared values ...
        curleftB = lout + totsamp;
        currightB = rout + totsamp;
        curwindow1 = window + totsamp;          // left window half
        curwindow2 = curwindow1 + sampleWindow; // right window half

        // add overlapping window halfs at the same time.
        i = cursamples % 8;
        while (i--)
            windowadd(fsqr(*curleftB++) + fsqr(*currightB++), *curwindow1++, *curwindow2++);
        i = cursamples / 8;
        while (i--)
        {   windowadd(fsqr(curleftB[0]) + fsqr(currightB[0]), curwindow1[0], curwindow2[0]);
            windowadd(fsqr(curleftB[1]) + fsqr(currightB[1]), curwindow1[1], curwindow2[1]);
            windowadd(fsqr(curleftB[2]) + fsqr(currightB[2]), curwindow1[2], curwindow2[2]);
            windowadd(fsqr(curleftB[3]) + fsqr(currightB[3]), curwindow1[3], curwindow2[3]);
            windowadd(fsqr(curleftB[4]) + fsqr(currightB[4]), curwindow1[4], curwindow2[4]);
            windowadd(fsqr(curleftB[5]) + fsqr(currightB[5]), curwindow1[5], curwindow2[5]);
            windowadd(fsqr(curleftB[6]) + fsqr(currightB[6]), curwindow1[6], curwindow2[6]);
            windowadd(fsqr(curleftB[7]) + fsqr(currightB[7]), curwindow1[7], curwindow2[7]);
            curleftB += 8;
            currightB += 8;
            curwindow1 += 8;
            curwindow2 += 8;
        }

        batchsamples -= cursamples;
        cursamplepos += cursamples;
        totsamp      += cursamples;
        if ( totsamp == sampleWindow ) {  // Get the Root Mean Square (RMS) for this set of samples
            // skip the first result, because it is only a half window.
            if (!first)
            {  // now store RMS
               int ival = 1806 + (int)((STEPS_per_dB * 10.) * log10(sum2 + 1.e-37)); // zero is bad
               if ( ival < 0 )
                  ival = 0;
                else if ( ival >= sizeof(A)/sizeof(*A) )
                  ival = sizeof(A)/sizeof(*A) - 1;
               A [ival]++;
            }
            first = 0;
            // move sum of left window half to right window half.
            sum2 = sum1;
            sum1 = 0.;
            // next half window
            memmove ( loutbuf , loutbuf  + totsamp, MAX_ORDER * sizeof *loutbuf );
            memmove ( routbuf , routbuf  + totsamp, MAX_ORDER * sizeof *routbuf );
            memmove ( lstepbuf, lstepbuf + totsamp, MAX_ORDER * sizeof *lstepbuf );
            memmove ( rstepbuf, rstepbuf + totsamp, MAX_ORDER * sizeof *rstepbuf );
            totsamp = 0;
        } else if ( totsamp > sampleWindow )  // somehow I really screwed up: Error in programming! Contact author about totsamp > sampleWindow
            return GAIN_ANALYSIS_ERROR;
    }
    if ( num_samples < MAX_ORDER ) {
        memmove ( linprebuf,                           linprebuf + num_samples, (MAX_ORDER-num_samples) * sizeof *linprebuf );
        memmove ( rinprebuf,                           rinprebuf + num_samples, (MAX_ORDER-num_samples) * sizeof *rinprebuf );
        memcpy  ( linprebuf + MAX_ORDER - num_samples, left_samples,            num_samples             * sizeof *linprebuf );
        memcpy  ( rinprebuf + MAX_ORDER - num_samples, right_samples,           num_samples             * sizeof *rinprebuf );
    }
    else {
        memcpy  ( linprebuf, left_samples  + num_samples - MAX_ORDER, MAX_ORDER * sizeof *linprebuf );
        memcpy  ( rinprebuf, right_samples + num_samples - MAX_ORDER, MAX_ORDER * sizeof *rinprebuf );
    }

    return GAIN_ANALYSIS_OK;
}


static Float_t
analyzeResult ( Uint32_t* Array, size_t len )
{
    Uint32_t  elems;
    Int32_t   upper;
    size_t    i;

    elems = 0;
    for ( i = 0; i < len; i++ )
        elems += Array[i];
    if ( elems == 0 )
        return GAIN_NOT_ENOUGH_SAMPLES;

    #ifdef DEBUG // write histogram data
    FILE* f = fopen("hist.out", "w");
    Uint32_t* ap = Array;
    #define HIST_BIN_SIZE 50
    for ( i = 0; i <= len - HIST_BIN_SIZE; i += HIST_BIN_SIZE )
    {   Uint32_t sum = 0;
        size_t j = HIST_BIN_SIZE;
        do sum += *ap++;
        while (--j);
        fprintf(f, "%g \t%g\n", (i + HIST_BIN_SIZE/2.) / STEPS_per_dB - pink_ref, (double)sum / elems);
    }
    fclose(f);
    #endif

    upper = (Int32_t) ceil (elems * (1. - rms_percentile));
    for ( i = len; i-- > 0; ) {
        if ( (upper -= Array[i]) <= 0 )
            break;
    }

    return (pink_ref - (Float_t)i / (Float_t)STEPS_per_dB);
}


Float_t
GetTitleGain ( void )
{
    unsigned int i;

    Float_t retval = analyzeResult ( A, sizeof(A)/sizeof(*A) );

    for ( i = 0; i < sizeof(A)/sizeof(*A); i++ ) {
        B[i] += A[i];
        A[i]  = 0;
    }

    #ifdef PROFILE
    printf("elapsed time: %f s\n", (clock() - t0) / (double)CLOCKS_PER_SEC);
    #endif

    resetGain();

    return retval;
}


Float_t
GetAlbumGain ( void )
{
    Float_t retval = analyzeResult ( B, sizeof(B)/sizeof(*B) );

/*    #ifdef PROFILE
    printf("elapsed time: %f s\n", (clock() - t0) / (double)CLOCKS_PER_SEC);
    #endif
*/
    return retval;
}

/* end of gain_analysis.c */


