#define WAVGAIN_VERSION "1.0.5"

/*
 *  wavgain.c - analyzes RIFF wav files, determines the perceived volume,
 *              and adjusts the volume of the file accordingly
 *
 *  Copyright (C) 2002 Glen Sawyer & Marcel Mueller
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  coding by Glen Sawyer (glensawyer@hotmail.com) 442 N 700 E, Provo, UT 84606 USA
 *    -- go ahead and laugh at me for my lousy coding skillz, I can take it :)
 *       Just do me a favor and let me know how you improve the code.
 *       Thanks.

 *  Ported to ANSI C (or at least somethingvery similar) by Marcel M�ller.
 *  /p won't work in this port, because there is no standard way to overwrite
 *  the last modified time stamp.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <sys/types.h>
//#include <sys/utime.h>
#include <sys/stat.h>
//#include <io.h>
#include <fcntl.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>

#ifndef max
#define max(a,b) ((a)<(b)?(b):(a))
#endif
#ifndef min
#define min(a,b) ((a)>(b)?(b):(a))
#endif

#include "gain_analysis.h"
#include "filefuncs.h"

#define BUFFERSIZE 1000000

static char buffer[BUFFERSIZE];

struct WAVheader
{  char      RIFF[4];
   uint32_t  BlkLen;
   char      WAVE[8];
   uint32_t  WAVLen;
   uint16_t  x1;
   uint16_t  Channels;
   uint32_t  SampRate;
   uint32_t  x2;
   uint16_t  BpSamp;
   uint16_t  Bits;
   char      data[4];
   uint32_t  DataLen;
};

/* Yes, yes, I know I should do something about these globals */

static const char* progname;

static bool ignoreClipWarning = 0;
static bool QuietMode = 0;
static bool InteractiveMode = 0;
static const char* defInput = "";
static int UsingTemp = 0;
static double dBGainMod = 0;
static double fGainMod = 1;
static bool applyRadio = 0;
static bool applyAlbum = 0;
static int databaseFormat = 0;
unsigned char maxAmpOnly = 0;

static int whichChannel = 0;
static double dBThres = 1; // changes less than 1 dB are ignored

static short int saveTime;
static bool timeOK = false;


static void Message(const char* fmt, ...)
// Print a informational message (if !QuietMode)
{  if (QuietMode)
      return;
   va_list va;
   va_start(va, fmt);
   vfprintf(stderr, fmt, va);
   va_end(va);
}

static
int scaleSamples(short *sp, int memsize, Float_t lscale, Float_t rscale)
{  memsize >>= 2;
   if (lscale > 1 || rscale > 1)
   {  // with overflow check
      int tmp;
      while (memsize--)
      {  tmp = (int)floor(sp[0] * lscale + rand() / (RAND_MAX + 1.));
         if (tmp > SHRT_MAX)
            tmp = SHRT_MAX;
          else if (tmp < SHRT_MIN)
            tmp = SHRT_MIN;
         sp[0] = (short)tmp;
         tmp = (int)floor(sp[1] * rscale + rand() / (RAND_MAX + 1.));
         if (tmp > SHRT_MAX)
            tmp = SHRT_MAX;
          else if (tmp < SHRT_MIN)
            tmp = SHRT_MIN;
         sp[1] = (short)tmp;
         sp += 2;
      }
   } else
   {  // without overflow check
      while (memsize--)
      {  sp[0] = (short)floor(sp[0] * lscale + rand() / (RAND_MAX + 1.));
         sp[1] = (short)floor(sp[1] * rscale + rand() / (RAND_MAX + 1.));
         sp += 2;
      }
   }
   return 0;
}

static int changeGainK(FILE* inf, FILE* outf, Float_t leftgainchange, Float_t rightgainchange)
{  // transfer header
   WAVheader Header;
   // read
   if (fread(&Header, sizeof Header, 1, inf) != 1)
   {  fprintf(stderr, "Failed to read header - Error %i.\n", errno);
      return 5;
   }

   if (  memcmp(Header.RIFF, "RIFF", sizeof Header.RIFF)
      || memcmp(Header.WAVE, "WAVEfmt ", sizeof Header.WAVE)
      || Header.WAVLen != 16
      || Header.BpSamp != 4
      || Header.Bits != 16
      || memcmp(Header.data, "data", sizeof Header.data) )
   {  fputs("Unrecognized file format (must be uncompressed 16bit stereo WAV).\n", stderr);
      return 6;
   }

   if (fwrite(&Header, sizeof Header, 1, outf) != 1)
   {  fputs("Failed to write WAV header\n", stderr);
      return 7;
   }

   // process data
   unsigned long rlen = Header.DataLen;
   while (rlen > 0)
   {  // read
      int n = fread(&buffer, 4, min(rlen, sizeof buffer)/4, inf);
      if (n == 0)
      {  fprintf(stderr, "Read error %i\n", errno);
         return 10;
      }
      n <<= 2;
      // scale samples
      if (scaleSamples((short*)buffer, n, leftgainchange, rightgainchange) != 0)
         return 11;
      // write back
      if (fwrite(&buffer, n, 1, outf) != 1)
      {  fprintf(stderr, "Write error %i\n", errno);
         return 12;
      }
      rlen -= n;

      Message("%ld of %ld bytes modified\r", Header.DataLen - rlen, Header.DataLen);
   }
   Message("                                       \r");

   return 0;
}

static
int changeGain(const char *filename, Float_t leftgainchange, Float_t rightgainchange)
{  fprintf(stderr,"Applying WAV gain change of %f/%f to %s...\n",leftgainchange,rightgainchange,filename);

   if (saveTime)
      fileTime(filename, storeTime);

   // input stream
   FILE* inf = fopen(filename,"rb");
   if (inf == NULL)
   {  fprintf(stderr, "Can't open %s for reading\n\n",filename);
      return 1;
   }

   // output stream (2 streams are significantly faster than a single one)
   const char* outfilename = UsingTemp
    ? strcat(strcpy((char*)malloc(strlen(filename)+5),filename),".tmp")
    : filename;
   FILE* outf = fopen(outfilename,"r+b");
   if (outf == NULL)
   {  fprintf(stderr, "Can't open %s for writing\n\n",outfilename);
      fclose(inf);
      return 2;
   }

   int r = changeGainK(inf, outf, leftgainchange, rightgainchange);

   fclose(inf);
   fclose(outf);

   if (saveTime && timeOK)
      fileTime(filename, setStoredTime);

   if (UsingTemp)
   {  if (r == 0)
      {  if (deleteFile(filename) == -1)
         {  fprintf(stderr, "Can't remove original file %s\n", filename);
            return 3;
         }
         if (moveFile(outfilename, filename) == -1)
         {  fprintf(stderr, "Can't open or rename %s to %s\n", outfilename, filename);
            return 4;
         }
      }
      free((char*)outfilename);
   } else if (r >= 10)
      fprintf(stderr, "Fatal I/O error: %s possibly left in an inconsistent state\n", filename);

   fprintf(stderr,"\n");
   return r;
}

static
int prepareSamples(const short *sp, int inmemsize, Float_t *outl, Float_t *outr, int* maxsamp)
{  inmemsize >>= 2;
   int maxs = *maxsamp;
   while (inmemsize--)
   {  *outl++ = sp[0];
      if (abs(sp[0]) > maxs)
         maxs = abs(sp[0]);
      *outr++ = sp[1];
      if (abs(sp[1]) > maxs)
         maxs = abs(sp[1]);
      sp += 2;
   }
   *maxsamp = maxs;
   return 0;
}

static void showVersion()
{  fprintf(stderr,"%s version %s\n",progname,WAVGAIN_VERSION);
}

static void errUsage()
{  showVersion();
   fprintf(stderr,"Usage: %s [options] <infile> [<infile 2> ...]\n\n",progname);
   fprintf(stderr,"  --use \"%s /?\" for a full list of options\n",progname);
}

static void fullUsage()
{  showVersion();
   fprintf(stderr,"Usage: %s [options] <infile> [<infile 2> ...]\n\n"
                  "options:\n"
                  "\t/v - show version number\n"
                  "\t/g <i>  - apply gain i to WAV without doing any analysis\n"
                  "\t/l0 <i> - apply gain i to channel 0 (left channel) of WAV\n"
                  "\t/l1 <i> - apply gain i to channel 1 (right channel) of WAV\n\n"
                  "\t/r - apply Radio gain automatically (all files set to equal loudness)\n"
                  "\t/a - apply Audiophile gain automatically (files are all from the same\n"
                  "\t              album: a single gain change is applied to all files, so\n"
                  "\t              their loudness relative to each other remains unchanged,\n"
                  "\t              but the average album loudness is normalized)\n"
                  "\t/m <f> - modify suggested WAV gain by factor f\n"
                  "\t/d <n> - modify suggested dB gain by double n\n"
                  "\t/c - ignore clipping warning when applying gain\n"
                  "\t/C<v> - force non-interactive mode, always use v as input\n"
                  "\t        /Cy is the same as /c. /Ca will always do a fail safe operation\n"
                  "\t        where the gain is never increased beyond the clipping limit.\n"
                  "\t/i - interactive mode: allways commit manually\n"
                  "\t/I<v> - same as /i but with the default input v in case you simply hit\n"
                  "\t        enter. /Ia is usually the best choice.\n"
                  "\t/o - output is a database-friendly tab-delimited list\n"
                  "\t/t - wavgain writes modified wav to temp file, then deletes original\n"
                  "\t     instead of modifying bytes in original file\n"
                  "\t/q - Quiet mode: no status messages\n"
                  "\t/p - Preserve original file timestamp\n"
                  "\t/x - Only find max. amplitude of wav\n"
                  "\t/? - show this message\n\n"
                  "If you specify /r and /a, only the second one will work\n"
                  "If you do not specify /c, the program will stop and ask before\n"
                  "     applying gain change to a file that might clip\n"
    ,progname);
}

/* Internal Tools (MM) */

static void PrintGainInfo(const char* name, double dBchange, int maxsample)
{  const char* type;
   if (name)
   {  type = "Radio";
      name = "";
   } else
   {  type = "Album";
      name = " for all files";
   }
   double factor = pow(10,dBchange/20);
   if (databaseFormat) {
      fprintf(stderr,"%s\t%f\t%f\t%d\n",name,factor,dBchange,maxsample);
      fflush(stderr);
   } else {
      fprintf(stderr,"Recommended \"%s\" dB change%s: %f\n",type,name,dBchange);
      fprintf(stderr,"Recommended \"%s\" WAV gain change%s: %f\n",type,name,factor);
      fprintf(stderr,"Max PCM sample at current gain: %d\n",maxsample);
      if (maxsample * factor > 32767.0)
         fprintf(stderr,"WARNING: some clipping may occur with this gain change!\n");
      fprintf(stderr,"\n");
}  }

static inline double factor2dB(double val)
{  return 20/log(10)*log(val);
}

static inline double dB2factor(double val)
{  return pow(10, val/20.);
}

static double CommitGainChange(const char* file, double val, double maxpcm)
{  if (file == NULL)
      file = "files";
   fflush(stderr);
   double clip = maxpcm * val / 32767.5;
   if (clip >= 1)
   {  fprintf(stderr,"\nWARNING: %s may clip with WAV gain change %f\n", file, val);
   }
   if (!InteractiveMode)
   {  if (fabs(factor2dB(val)) < dBThres)
      {  fprintf(stderr,"\nNo changes to %s are necessary\n", file);
         return 1;
      }
      if (clip < 1 || ignoreClipWarning)
         return val;
   }
   // interactive
   double aval = val;
   if (val > 1 && clip > 1)
      aval /= clip;
   do
   {  fprintf(stderr,"Make change? [y(%f)|n(1)|a(%f)|<value>|<dB>d]:", val, aval);
      fflush(stderr);
      char buf[10] = {0};
      scanf("%*[ \t]");
      scanf("%9[^ \t\n]", buf);
      scanf("%*[ \t]");
      if (buf[0] == 0)
      {  if (defInput[0] == 0)
            continue;
         strncpy(buf, defInput, 9);
         puts(buf);
      }
      if (buf[1] == 0)
      switch (toupper(buf[0]))
      {case 'Y':
         return val;
       case 'N':
         return 1;
       case 'A':
         return aval;
      }
      bool isdB = false;
      size_t len = strlen(buf);
      if (toupper(buf[len-1]) == 'D')
      {  isdB = true;
         buf[--len] = 0;
      }
      double nval;
      size_t n;
      if (sscanf(buf, "%lf%n", &nval, (int*)&n) == 1 && n == strlen(buf))
         return isdB ? dB2factor(nval) : nval;
   } while (true);
}


static int first = 1;

static int AnalyzeFile(const char* name, int* maxsamp)
{  int nchan = 2; // stereo only so far
   static Float_t lsamples[BUFFERSIZE/4];
   static Float_t rsamples[BUFFERSIZE/4];

   if (!databaseFormat)
      fprintf(stderr,"%s\n",name);

   FILE* inf = fopen(name,"rb");
   if (inf == NULL)
   {  fprintf(stderr, "Can't open %s for reading\n\n",name);
      return 1;
   }
   // analyze file

   WAVheader Header;

   // WAV header
   if (fread(&Header, sizeof Header, 1, inf) != 1)
   {  fprintf(stderr, "Failed to read header - Error %i.\n", errno);
      return 2;
   }

   if (  memcmp(Header.RIFF, "RIFF", sizeof Header.RIFF)
      || memcmp(Header.WAVE, "WAVEfmt ", sizeof Header.WAVE)
      || Header.WAVLen != 16
      || Header.BpSamp != 4
      || Header.Bits != 16
      || memcmp(Header.data, "data", sizeof Header.data) )
   {  fclose(inf);
      fputs("Unrecognized file format (must be uncompressed 16bit stereo WAV).", stderr);
      return 3;
   }

   *maxsamp = 0;

   if (!maxAmpOnly) {
      if (first) {
         InitGainAnalysis(Header.SampRate);
         first = 0;
      } else
         ResetSampleFrequency(Header.SampRate);
   }

   // analyze content
   unsigned long rlen = Header.DataLen;
   while (rlen > 0)
   {  int n = fread(&buffer, 4, min(rlen, sizeof buffer)/4, inf);
      if (n == 0)
      {  fprintf(stderr, "Read error %i\n", errno);
         fclose(inf);
         return 4;
      }

      if (prepareSamples((short*)buffer,n<<2,lsamples,rsamples,maxsamp) == 0) {
         if (!maxAmpOnly) {
            if (AnalyzeSamples(lsamples,rsamples,n,nchan) == GAIN_ANALYSIS_ERROR) {
               fprintf(stderr,"Error analyzing further samples (max time reached)          \n");
               fclose(inf);
               return 5;
            }
         }
      }

      rlen -= n<<2;

      Message("%ld of %ld bytes analyzed\r", Header.DataLen - rlen, Header.DataLen);
   }
   Message("                                       \r");

   fclose(inf);

   return 0;
}

/* Internal Tools end */

// coomand line
static void Fatal(int no, const char* fmt, ...)
{  va_list va;
   va_start(va, fmt);
   vfprintf(stderr, fmt, va);
   va_end(va);
   errUsage();
   exit(no);
}

static double FloatOpt(const char**& src, const char*& par)
{  double r;
   size_t n = 0;
   sscanf(par, "%lf%zn", &r, &n);
   if (n)
   {  // inline option
      par += n;
      return r;
   }
   // try next param
   if (*src == NULL)
      Fatal(42, "Option %c expects a floating point parameter.\n", par[-1]);
   sscanf(*src, "%lf%zn", &r, &n);
   if (n != strlen(*src))
      Fatal(42, "Option %c expects a floating point parameter, found %s.\n", par[-1], *src);
   ++src;
   return r;
}

static int IntOpt(const char**& src, const char*& par)
{  int r;
   size_t n = 0;
   sscanf(par, "%i%zn", &r, &n);
   if (n)
   {  // inline option
      par += n;
      return r;
   }
   // try next param
   if (*src == NULL)
      Fatal(42, "Option %c expects an integer parameter.\n", par[-1]);
   sscanf(*src, "%i%zn", &r, &n);
   if (n != strlen(*src))
      Fatal(42, "Option %c expects an integer parameter, found %s.\n", par[-1], *src);
   ++src;
   return r;
}
static int IntOpt(const char**& src, const char*& par, int minval, int maxval)
{  int r = IntOpt(src, par);
   if (r < minval || r > maxval)
      Fatal(42, "Parameter %i out of range [%i,%i].\n", r, minval, maxval);
   return r;
}

int main(int argc, const char **argv) {
   double dBchange;
   int mainloop;
   int *maxsample;
   int *fileok;
   int directGain = 0;
   int directSingleChannelGain = 0;
   Float_t directGainVal = 0;

   saveTime = 0;

   progname = *argv++;

   if (argc < 2)
      errUsage();

   // parse command line
   const char** src = argv;   // to parse the options
   const char** dst = argv;   // to store the files

   while (*src)
   {  const char* param = *src++;
      if (param[0] == '/')
      {  ++param;
         while (param && *param)
         {  switch(*param++)
            {case 'a':
               applyRadio = false;
               applyAlbum = true;
               break;
             case 'C':
               defInput = param;
               ignoreClipWarning = true;
               break;
             case 'c':
               ignoreClipWarning = true;
               defInput = "y";
               break;
             case 'd':
               dBGainMod = FloatOpt(src, param);
               break;
             case 'g':
               directGain = true;
               directSingleChannelGain = false;
               directGainVal = FloatOpt(src, param);
               break;
             case 'I':
               defInput = param;
               param = NULL; // ate
             case 'i':
               InteractiveMode = true;
               break;
             case 'l':
               directSingleChannelGain = true;
               directGain = false;
               whichChannel = IntOpt(src, param, 0,1);
               break;
             case 'm':
               fGainMod = FloatOpt(src, param);
               break;
             case 'o':
               databaseFormat = true;
               break;
             case 'p':
               saveTime = true;
               break;
             case 'q':
               QuietMode = true;
               break;
             case 'r':
               applyRadio = true;
               applyAlbum = false;
               break;
             case 't':
               UsingTemp = true;
               break;
             case 'x':
               maxAmpOnly = true;
               break;
             case 'h':
             case '?':
               fullUsage();

             case 'v':
               showVersion();
               exit(0);
             default:
               Fatal(44, "I don't recognize option %s\n", param[-1]);
            }
         } // other options...
      } else
         // no option! => filename
         *dst++ = param;
   }
   const int numFiles = dst - argv;
   if (numFiles == 0)
      Fatal(10, "Nothing to do.\n");

   maxsample = (int*)malloc(sizeof(int) * numFiles); /* new double[argc]; */
   fileok = (int*)malloc(sizeof(int) * numFiles); /* new int[argc]; */

   if (databaseFormat)
      fprintf(stdout,"File\tMP3 gain\tdB gain\tMax Amplitude\n");

   // main loop
   for (mainloop = 0; mainloop < numFiles; ++mainloop)
   {  fileok[mainloop] = 0;

      if (directSingleChannelGain)
      {  if (whichChannel) /* do right channel */
            changeGain(argv[mainloop],0,directGainVal);
          else  /* do left channel */
            changeGain(argv[mainloop],directGainVal,0);
      } else if (directGain)
         changeGain(argv[mainloop],directGainVal,directGainVal);
      // analyze !!
       else if ((fileok[mainloop] = !AnalyzeFile(argv[mainloop], &maxsample[mainloop])) != 0)
      {  if (maxAmpOnly)
            dBchange = 0;
          else
         {  dBchange = GetTitleGain();
            if (dBchange == GAIN_NOT_ENOUGH_SAMPLES)
            {  fprintf(stderr,"Not enough samples in WAV files to do analysis\n\n");
               continue;
         }  }
         dBchange += dBGainMod + factor2dB(fGainMod);

         if (InteractiveMode || (!applyRadio && !applyAlbum))
            PrintGainInfo(argv[mainloop],dBchange,maxsample[mainloop]);
         if (applyRadio)
         {  first = true; /* don't keep track of Audiophile gain */
            double factor = CommitGainChange(argv[mainloop], dB2factor(dBchange), maxsample[mainloop]);
            if (factor != 1)
               // do radio gain !!!
               changeGain(argv[mainloop],factor,factor);
         }
         fflush(stderr);
      }
   }

   if (!applyRadio)
   {  if (maxAmpOnly)
         dBchange = 0;
       else
      {  dBchange = GetAlbumGain();
         if (dBchange == GAIN_NOT_ENOUGH_SAMPLES)
         {  fprintf(stderr,"Not enough samples in WAV files to do analysis\n\n");
            goto end;
      }  }
      dBchange += dBGainMod + factor2dB(fGainMod);

      int maxmaxsample = 0;
      for (mainloop = 0; mainloop < numFiles; ++mainloop)
         if (fileok[mainloop] && maxsample[mainloop] > maxmaxsample)
            maxmaxsample = maxsample[mainloop];

      if (InteractiveMode || !applyAlbum)
         PrintGainInfo(NULL, dBchange, maxmaxsample);
      if (applyAlbum)
      {  double factor = CommitGainChange(NULL, dB2factor(dBchange), maxmaxsample);
         if (factor != 1)
            // do album gain !!!
            for (mainloop = 0; mainloop < numFiles; ++mainloop)
               if (fileok[mainloop])
                  // do album gain !!!
                  changeGain(argv[mainloop],factor,factor);
      }
   }
 end:
   free(maxsample);
   free(fileok);

   return 0;
}
